---
title: Aider blog
layout: default
nav_order: 1000
---

<h1>Aider blog</h1>

<div class="posts">
    {% for post in site.posts %}
    {% unless post.draft %}
    <article class="post">
      <h2><a href="{{ site.baseurl }}{{ post.url }}">{{ post.title }}</a></h2>
      <div class="entry">
        <div class="post-content">
          {% if post.excerpt %}
          {{ post.excerpt }}
          {% else %}
          {{ post.content | strip_html | truncatewords: 100 }}...
          {% endif %}
          {% if post.highlight_image %}
          <a href="{{ site.baseurl }}{{ post.url }}">
              <div class="post-highlight">
                  <img src="{{ site.baseurl }}{{ post.highlight_image }}" alt="Highlight Image">
              </div>
          </a>
          {% endif %}
        </div>
      </div>

      <p class="post-date">
          {{ post.date | date: "%b %-d, %Y" | upcase }}
      </p>
    </article>
    {% endunless %}
    {% endfor %}
</div>
