<footer class="site-footer">
    Aider is AI pair programming in your terminal.
    Aider is on
    <a href="https://github.com/Aider-AI/aider">GitHub</a>
    and
    <a href="https://discord.gg/Tv2uQnR88V">Discord</a>.
</footer>
